import copy
import random
import numpy as np
import math
from modelestimator._bw_estimator.bw_estimator import bw_estimator

def _resample_columns(sequence_list):
    return_list = [""] * len(sequence_list)
    SEQUENCE_LENGTH = len(sequence_list[0])

    for _ in range(SEQUENCE_LENGTH):
        RANDOM_INDEX = random.randint(0, SEQUENCE_LENGTH - 1)

        for SEQUENCE_INDEX, SEQUENCE in enumerate(sequence_list):
            SEQUENCE_ELEMENT = SEQUENCE[RANDOM_INDEX]
            return_list[SEQUENCE_INDEX] = return_list[SEQUENCE_INDEX] + SEQUENCE_ELEMENT

    return return_list

def _calculate_bw_for_resamplings(FORMAT, RESAMPLINGS, THRESHOLD, MULTIALIGNMENT):
    q_list = []
    eq_list = []
    number_of_times_bw_estimator_failed = 0

    for _ in range(RESAMPLINGS):
        MULTIALIGNMENT = copy.deepcopy(MULTIALIGNMENT)
        MULTIALIGNMENT = _resample_columns(MULTIALIGNMENT)
        MULTIALIGNMENT_LIST = [MULTIALIGNMENT]

        try:
            Q, EQ = bw_estimator(FORMAT, THRESHOLD, MULTIALIGNMENT_LIST)
            q_list.append(Q)
            eq_list.append(EQ)
        except:
            number_of_times_bw_estimator_failed +=1

    FAILED_PERCENTAGE = number_of_times_bw_estimator_failed / RESAMPLINGS
    return q_list, FAILED_PERCENTAGE

def q_diff_mean(REFERENCE_Q, RESAMPLED_Q_LIST):
    Q_DIFF_NORM_LIST = []

    for Q in RESAMPLED_Q_LIST:
        Q_DIFF = REFERENCE_Q - Q
        Q_DIFF_NORM = np.linalg.norm(Q_DIFF)
        Q_DIFF_NORM_LIST.append(Q_DIFF_NORM)

    Q_DIFF_MEAN = np.mean(Q_DIFF_NORM_LIST)

    return Q_DIFF_MEAN

#   Interface
def bootstraper(FORMAT, RESAMPLINGS, THRESHOLD, MULTIALIGNMENT):
    MULTIALIGNMENT_LIST = [MULTIALIGNMENT]
    REFERENCE_Q,_ = bw_estimator(FORMAT, THRESHOLD, MULTIALIGNMENT_LIST)

    RESAMPLED_Q_LIST, FAILED_PERCENTAGE = _calculate_bw_for_resamplings(FORMAT, RESAMPLINGS, THRESHOLD, MULTIALIGNMENT)
    Q_DIFF_MEAN = q_diff_mean(REFERENCE_Q, RESAMPLED_Q_LIST)
    Q_DIFF_MEAN *= 10000

    return Q_DIFF_MEAN, FAILED_PERCENTAGE


import sys
from modelestimator._parse_arguments import parse_arguments
from modelestimator._controller import controller

USAGE_STRING ="""Usage: python -m modelestimator <format> <options> infiles

<format> should be either FASTA, STOCKHOLM or PHYLIP
Output is a rate matrix and residue distribution vector.
        
Options:
    -threshold or -t <f> Stop when consecutive iterations do not change by more
                     than <f>. Default is 0.001.
    -bootstrap or -b <r> Perform bootstrapping on multialignment with <r> resamplings.
                         Only one infile should be given in this mode. Returns
                         bootstrap norm.

Example usage:
    modelestimator fasta -t 0.001 file1.fa file2.fa file3.fa
    modelestimator fasta -b 200 file.fa"""

def main():
    ARGUMENT_LIST = sys.argv[1:]
      
    if len(ARGUMENT_LIST) == 0:
        print(USAGE_STRING)
        exit()
            
    try:
        FORMAT, BOOTSTRAP, RESAMPLINGS, THRESHOLD, FILE_NAMES = parse_arguments(ARGUMENT_LIST)
    except Exception as e:
        print("Wrong format on input\n")
        print(USAGE_STRING)
        exit()
        
    try:
        OUTPUT_STRING = controller(FORMAT, BOOTSTRAP, RESAMPLINGS, THRESHOLD, FILE_NAMES)
    except Exception as e:
        print("Error: ", e)
        exit()

    print(OUTPUT_STRING)


from modelestimator._bw_estimator.bw_estimator import bw_estimator
from modelestimator._handle_input.handle_input_file import handle_input_file
from modelestimator._bootstraper.bootstraper import bootstraper

def controller(FORMAT, BOOTSTRAP, RESAMPLINGS, threshold, FILE_NAME_LIST):
    MULTIALIGNMENT_LIST = []
    
    for FILE in FILE_NAME_LIST:
        MULTIALIGNMENT = handle_input_file(FILE, FORMAT)
        MULTIALIGNMENT_LIST.append(MULTIALIGNMENT) 
        
    if threshold == None:
        threshold = 0.001
    
    if BOOTSTRAP:
        MULTIALIGNMENT = MULTIALIGNMENT_LIST[0]
        BOOTSTRAP_NORM,_ = bootstraper(FORMAT, RESAMPLINGS, threshold, MULTIALIGNMENT)
        OUTPUT_STRING = "Bootstrap norm = " + str(BOOTSTRAP_NORM)
    else:
        Q, EQ = bw_estimator(FORMAT, threshold, MULTIALIGNMENT_LIST)
        OUTPUT_STRING = "Q =\n" + str(Q) + "\nEQ =\n" + str(EQ)   

    return OUTPUT_STRING


import sys
import os
import io
from contextlib import redirect_stdout
import numpy as np
from modelestimator.__main__ import main

def test_integration_1():
    #   Load REFERENCE_OUTPUT_STRING
    reference_output_string_path = "tests\\test_integration_1\\reference-output-string.txt"
    reference_output_string_path = os.path.join(sys.path[0], reference_output_string_path)

    with open(reference_output_string_path) as text_file:
            REFERENCE_OUTPUT_STRING = text_file.read()
    
    #   Calculate OUTPUT_STRING
    reference_input_string_path = "\\tests\\test_integration_1\\test_integration_1_20seqs_1000long_50pam.fa"
    reference_input_string_path = sys.path[0] + reference_input_string_path
    sys.argv = ["-", "fasta", reference_input_string_path]
    
    f = io.StringIO()
    with redirect_stdout(f):
        main()
    
    OUTPUT_STRING = f.getvalue()

    assert(OUTPUT_STRING == REFERENCE_OUTPUT_STRING)


def parse_arguments(argument_list):
    FORMAT = argument_list.pop(0)
    FORMAT = FORMAT.lower()

    if FORMAT not in ['fasta', 'stockholm', 'phylip']:
        raise Exception()

    #   Options
    bootstrap = False
    threshold = None
    resamplings = None

    while argument_list[0][0] == "-":
        if argument_list[0][1] == "t" or argument_list[0][1] == "threshold":
            argument_list.pop(0)
            threshold = float(argument_list.pop(0))
            if not(threshold > 0 and threshold < 1):
                raise Exception()
        
        elif argument_list[0][1] == "b" or argument_list[0][1] == "bootstrap":            
            argument_list.pop(0)
            bootstrap = True

            resamplings = int(argument_list.pop(0))
            if not(resamplings > 0):
                raise Exception()


    FILE_LIST = argument_list
    if (len(FILE_LIST) == 0) or (bootstrap and len(FILE_LIST) > 1):
        raise Exception()

    return FORMAT, bootstrap, resamplings, threshold, FILE_LIST